@@ -1,3 +1,4 @@
+from pathlib import Path
 from torch import nn
 from transformers.generation_logits_process import LogitsProcessorList
 from transformers.generation_stopping_criteria import (
@@ -14,7 +15,7 @@ import torch
 import torch.distributed as dist
 import warnings
 
-from htools import tolist
+from htools import tolist, save, spacer, add_docstring
 
 
 logger = logging.get_logger(__name__)
@@ -563,7 +564,9 @@ def live_generate(
     Most of these parameters are explained in more detail in [this blog
     post](https://huggingface.co/blog/how-to-generate).
     Parameters:
-        inputs (`torch.Tensor` of varying shape depending on the modality, *optional*):
+        inputs (`torch.Tensor` of varying shape depending on the modality,
+                *optional*. HDM update: added support for passing in str. Only
+                1 prompt can be provided at a time (at least currently).):
             The sequence used as a prompt for the generation or as model inputs to the encoder. If `None` the
             method initializes it with `bos_token_id` and a batch size of 1. For decoder-only models `inputs`
             should of in the format of `input_ids`. For encoder-decoder models *inputs* can represent any of
@@ -712,8 +715,11 @@ def live_generate(
     num_beam_groups = 1
     prefix_allowed_tokens_fn = None
 
-    if isinstance(inputs, (str, list)):
+    if isinstance(inputs, str):
         inputs = tokenizer(inputs, return_tensors='pt').input_ids
+    if isinstance(inputs, list) or inputs.shape[0] > 1:
+        raise NotImplementedError('We currently only support live generation '
+                                  'for one input prompt at a time.')
 
     # 1. Set generation parameters if not already defined
     bos_token_id = (bos_token_id if bos_token_id is not None
@@ -1005,4 +1011,138 @@ def truncate_if_necessary(tokenizer, text, max_new_tokens,
         'truncated_length': new_len,
         'truncated_length_words': new_len_words,
     }
+    return res
+
+
+def word_to_ids(tokenizer, word):
+    """Convert a word (str) to a list of word indices (ints). Using this in
+    `generate` lets user pass in list of strings for words to skip/force
+    instead of list[list[int]].
+
+    Parameters
+    ----------
+    tokenizer: transformers Tokenizer
+    word: str
+
+    Returns
+    -------
+    list[int]
+    """
+    return tokenizer.convert_tokens_to_ids(
+        tokenizer.tokenize(word)
+    )
+
+
+@add_docstring(live_generate)
+def generate(
+        model,
+        tokenizer,
+        prompt,
+        out_dir=None,
+        mode_pre='a',
+        repetition_penalty=1.2,
+        early_stopping=False,
+        do_sample=True,
+        temperature=0.7,
+        top_p=.99,
+        **kwargs
+):
+    """
+    Parameters
+    ----------
+    model
+    tokenizer
+    prompt: str
+        Notice that you must pass in a single prompt (str), not a list of
+        prompts.
+    out_dir: str or Path or None
+        If not None, a file will be saved in this directory with the generated
+        output(s).
+    mode_pre: str
+        'a' or 'w'. If out_dir is provided, this determines what mode to use
+        when writing to that file (append or write). Append is the default
+        because the immediate use case is to use this in a callback that will
+        be executing repeatedly during training.
+    repetition_penalty
+    early_stopping
+    do_sample
+    temperature
+    top_p
+    kwargs: any
+        Additional kwargs for model.generate().
+
+    Returns
+    -------
+    list[dict]: Each dict contains data for one generated sequence (we can use
+    `num_return_sequences` arg to request > 1). Each dict has keys:
+        tokens: list[str]
+            List of generated tokens.
+        probs: list[dict[str, float]]
+            One dict for each time step in generation. Each dict maps token to
+            its corresponding probability. Contains 5 most likely tokens and
+            the selected token (when not in greedy mode, this means you can get
+            either 5 or 6 probs in each dict).
+        text: str
+            The de-tokenized generated sequence as a single str.
+        full_text: str
+            Same as text but with input prompt too. Prompt and generation
+            are distinguished by sub-headers PROMPT and RESPONSE. Useful when
+            saving output to a file (in fact, when out_dir is provided, this
+            is basically what we save, with the addition of some spacers and
+            numbering).
+    """
+    if not isinstance(prompt, str):
+        raise TypeError(f'Expected prompt to be str, got {type(prompt)}.')
+    if 'inputs' in kwargs:
+        del kwargs['inputs']
+        warnings.warn('You should provide a str/list `prompt` instead of the '
+                      'huggingface `inputs` arg.')
+
+    n = kwargs.get('num_return_sequences', 1)
+    kwargs.update(
+        repetition_penalty=repetition_penalty,
+        early_stopping=early_stopping,
+        do_sample=do_sample,
+        temperature=temperature,
+        top_p=top_p
+    )
+
+    # Make it easier to provide words to skip/include. Lets us pass in lists
+    # of strings instead of lists of lists of token IDs.
+    # HF does not accept empty list, has to be None if falsy.
+    kwargs['bad_words_ids'] = [word_to_ids(tokenizer, t) for t in
+                               kwargs.pop('skip_words', [])] or None
+    kwargs['force_words_ids'] = [word_to_ids(tokenizer, t) for t in
+                                 kwargs.pop('force_words', [])] or None
+    if kwargs['force_words_ids']:
+        if kwargs.get('do_sample', True):
+            warnings.warn('Auto-setting do_sample=False because force_words '
+                          'were passed in.')
+            kwargs['do_sample'] = False
+        if kwargs.get('num_beams', 1) <= 1:
+            warnings.warn('Auto-setting num_beams=2 because force_words were '
+                          'passed in.')
+            kwargs['num_beams'] = 2
+
+    # res[i] is a dict of results for the i'th generated sequence.
+    res = [{k: [] for k in ('tokens', 'probs')} for _ in range(n)]
+    for word, probs, is_alive in live_generate(model, tokenizer, prompt,
+                                               **kwargs):
+        for cur_word, cur_probs, cur_alive, cur_res in zip(word, probs,
+                                                           is_alive, res):
+            if cur_alive:
+                cur_res['tokens'].append(cur_word)
+                cur_res['probs'].append(cur_probs)
+    for i in range(len(res)):
+        res[i]['text'] = tokenizer.convert_tokens_to_string(res[i]['tokens'])
+        res[i]['full_text'] = ('# PROMPT\n' + prompt + '\n\n# RESPONSE\n'
+                               + res[i]['text'])
+    if out_dir:
+        out_dir = Path(out_dir)
+        save(
+            ''.join(f"{i}.\n{row['full_text']}{spacer()}\n"
+                    for i, row in enumerate(res, 1)),
+            out_dir/'generated_text_samples.txt',
+            mode_pre=mode_pre
+        )
     return res
\ No newline at end of file
